import hashlib
import re
import sys
from concurrent.futures import ThreadPoolExecutor, as_completed
from datetime import datetime, timedelta
from io import StringIO
from operator import itemgetter
from zoneinfo import ZoneInfo

import pandas as pd
import requests
from tqdm.auto import tqdm

from minitrade.datasource.base import QuoteSource
from minitrade.utils.mtdb import MTDB


class CboeFuturesQuoteSource(QuoteSource):
    '''CBOEFuturesQuoteSource retrieves historical futures data from [CBOE website](https://www.cboe.com/us/futures/market_statistics/historical_data/).

Accepted symbol format:
- CBOE futures: VX, VXF4, VXF24, VX@2024-01-17, ...

Daily bar:
- The daily OHLCV data returns historical data up to T-1.

Minute bar:
- Not available

Spot price:
- Not available

Limitations:
- Data is only available for monthly futures.
    '''

    def __init__(self, cf_method='backward', roll_day=-1):
        """
        Initialize a CboeFuturesQuoteSource object.

        Args:
            cf_method (str, optional): The method used for calculating continunous futures. Defaults to 'backward'.
            roll_day (int, optional): Day offset before expiration to roll to the next contract, should be negative. Defaults to -1.
        """
        super().__init__()
        self.cf_method = cf_method
        self.roll_day = roll_day

    def _download_futures_history(self, futures_root, num_workers=8):
        index_url = 'https://www.cboe.com/us/futures/market_statistics/historical_data/product/list/{}/'
        data_url = 'https://cdn.cboe.com/{}'
        key_fields = ['product_display', 'expire_date', 'contract_dt', 'futures_root', 'duration_type', 'hash']

        def fetch_data(file):
            now = datetime.now(ZoneInfo('US/Central'))
            try:
                cache = MTDB.get_one('CboeFuturesCache', 'hash', file['hash'])
                if cache['freeze'] or cache['update_time'] > now.replace(hour=0, minute=0, second=0, microsecond=0):
                    return itemgetter(*key_fields)(file), pd.read_csv(StringIO(cache['data']))
            except Exception:
                pass
            data = requests.get(data_url.format(file['path'])).text
            freeze = datetime.strptime(file['expire_date'], '%Y-%m-%d').date() < now.date() - timedelta(days=3)
            MTDB.save('CboeFuturesCache', file | {'data': data, 'freeze': freeze,
                      'update_time': now}, on_conflict='update')
            return itemgetter(*key_fields)(file), pd.read_csv(StringIO(data))

        def hash(file):
            return hashlib.md5(','.join(file.values()).encode()).hexdigest()

        data_files = [x for year in requests.get(index_url.format(futures_root)).json().values() for x in year]
        if not data_files:
            raise ValueError(f'No data found for {futures_root}')
        data_files = [file | {'hash': hash(file)} for file in data_files]
        with ThreadPoolExecutor(max_workers=num_workers) as executor:
            data = {itemgetter(*key_fields)(file): None for file in data_files}
            futures = [executor.submit(fetch_data, file) for file in data_files]
            for future in tqdm(
                    as_completed(futures),
                    total=len(data_files),
                    desc='Downloading data', unit='file', leave=False, file=sys.stdout):  # do not clutter the stderr log
                key, df = future.result()
                data[key] = df

        df = pd.concat(data).reset_index().drop(
            columns=['level_6']).rename(
            columns=dict(zip(['level_0', 'level_1', 'level_2', 'level_3', 'level_4', 'level_5'], key_fields))).sort_values(
            ['Trade Date', 'expire_date'])
        df = df[df['duration_type'] == 'M'].copy().rename(columns={'Trade Date': 'date', 'Total Volume': 'Volume'})
        return df

    def _continuous_front_month_futures(self, df, cf_method, roll_day=-5):
        """
        Calculates the continuous near-month futures prices based on the provided dataframe.

        Args:
            df (pd.DataFrame): The dataframe containing historical futures data.
            method (str | None): The method to use for adjusting the prices. Options are 'proportional' and 'backward', None for no adjustment.
            roll_day (int): Day offset before expiration to roll to the next contract, should be negative.

        Returns:
            pd.DataFrame: The dataframe with the continuous near-month futures prices.

        """
        df = self._futures_term_structure(df)
        ohlc = ['Open', 'High', 'Low', 'Close']
        out = df['M0'].copy()
        if cf_method:
            if roll_day >= 0:
                raise ValueError(f'Invalid roll day offset {roll_day}')
            expire_index = (df['M0', 'expire_date'] == df.index)
            roll_index = expire_index.shift(roll_day, fill_value=False)
            roll_fill = pd.concat([expire_index.shift(i) for i in range(roll_day+1, 1)],
                                  axis=1).any(axis=1).reindex(df.index)
            out[roll_fill] = df['M1'][roll_fill]
            if cf_method == 'proportional':
                adjust_ratio = df[roll_index]['M1', 'Close'] / df[roll_index]['M0', 'Close']
                adjust_ratio_fill = adjust_ratio.iloc[::-1].cumprod().iloc[::-1].reindex(df.index).bfill().fillna(1)
                out[ohlc] = out[ohlc].mul(adjust_ratio_fill, axis=0)
            elif cf_method == 'backward':
                adjust_delta = df[roll_index]['M1', 'Close'] - df[roll_index]['M0', 'Close']
                adjust_delta_fill = adjust_delta.iloc[::-1].cumsum().iloc[::-1].reindex(df.index).bfill().fillna(0)
                out[ohlc] = out[ohlc].add(adjust_delta_fill, axis=0)
            else:
                raise ValueError(f'Invalid adjust method {cf_method}')
        return out

    def _futures_term_structure(self, df):
        """
        Calculate the futures term structure based on the given DataFrame.

        Args:
            df (pd.DataFrame): The input DataFrame containing futures data.

        Returns:
            pd.DataFrame: The DataFrame representing the futures term structure.
        """
        df = df.copy()
        df['Month'] = df.groupby('date').cumcount()
        ohlcve = ['Open', 'High', 'Low', 'Close', 'Volume', 'expire_date']
        df = df.pivot_table(
            index='date', columns='Month', values=ohlcve, aggfunc='first')[ohlcve].swaplevel(
            axis=1).sort_index(axis=1)
        df.columns = df.columns.map(lambda x: (f'M{x[0]}', x[1]))
        df = df.reindex(zip(df.columns.get_level_values(0), ohlcve*len(df.columns.get_level_values(0))), axis=1)
        return df

    def _format_ticker(self, ticker):
        month_codes = 'FGHJKMNQUVXZ'
        try:
            root, month, year, root2, root3, expire_date = re.match(
                r'^([A-Z]+\d*)([FGHJKMNQUVXZ])(\d+)$|^([A-Z]+\d*)$|^([A-Z]+\d*)\@(\d{4}-\d{2}-\d{2})$', ticker).groups()
            root = root or root2 or root3
            month = month_codes.index(month) + 1 if month else None
            if year:
                year_now = datetime.now(ZoneInfo('US/Central')).year
                year = int(year)
                if year < 10:
                    year = year_now // 10 * 10 + year
                elif year < 100:
                    year = year_now // 100 * 100 + year
                else:
                    raise ValueError(f'Invalid year {year}')
            expire_date = expire_date or (f'{year}-{month:02d}' if year and month else None)
            return root, expire_date
        except Exception:
            raise ValueError(f'Invalid ticker {ticker}')

    def _ticker_timezone(self, ticker):
        return 'US/Central'

    def _ticker_calendar(self, ticker):
        return 'CBOE_Futures'

    def _daily_bar(self, ticker, start, end):
        root, expire_date = self._format_ticker(ticker)
        df = self._download_futures_history(root)
        if expire_date:
            a = df[df['expire_date'] == expire_date]
            df = df[df['expire_date'].str.slice(0, 7) == expire_date] if a.empty else a
            if df.empty:
                raise RuntimeError(f'No data found for {ticker}')
            df.set_index('date', inplace=True)
        else:
            df = self._continuous_front_month_futures(df, self.cf_method, self.roll_day)
        df = df.loc[start:]
        if end:
            df = df.loc[:end]
        df.index = pd.to_datetime(df.index)
        df.index.rename('Date', inplace=True)
        df = df[['Open', 'High', 'Low', 'Close', 'Volume', 'expire_date']]
        return df

    def _minute_bar(self, ticker, start, end, interval):
        raise NotImplementedError()

    def _spot(self, tickers):
        raise NotImplementedError()

    def term_structure(self, ticker, start=None, end=None):
        """
        Retrieves the term structure data for a given futures ticker.

        Parameters:
        ticker (str): The futures ticker symbol.
        start (str, optional): The start date of the term structure data. Defaults to None.
        end (str, optional): The end date of the term structure data. Defaults to None.

        Returns:
        pandas.DataFrame: The term structure data as a DataFrame.
        """
        root, _ = self._format_ticker(ticker)
        df = self._futures_term_structure(self._download_futures_history(root))
        df = df.xs('Close', axis=1, level=1)
        if start:
            df = df.loc[start:]
        if end:
            df = df.loc[:end]
        df.index = pd.to_datetime(df.index)
        df.index.rename('Date', inplace=True)
        return df



import datetime
import json
import logging
import os
import sqlite3
import sys
import threading
from dataclasses import asdict
from io import StringIO
from typing import Any

import pandas as pd
from nanoid import generate
from pypika import Order, Query, Table

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def adapt_json(val):
    """Adapt object to Json string."""
    return json.dumps(val, default=str) if val is not None else None


def adapt_dataframe(val: pd.DataFrame | pd.Series | None) -> str | None:
    return val.to_csv() if val is not None else None


def adapt_bytes(val) -> str | None:
    return val.decode() if val is not None else None


def convert_date(val):
    """Convert ISO 8601 date to datetime.date object."""
    return datetime.date.fromisoformat(val.decode())


def convert_datetime(val):
    """Convert ISO 8601 datetime to datetime.datetime object."""
    return datetime.datetime.fromisoformat(val.decode())


def convert_timestamp(val):
    """Convert Unix epoch timestamp to datetime.datetime object."""
    return datetime.datetime.fromtimestamp(int(val))


def convert_dataframe(csv: bytes) -> pd.DataFrame | pd.Series:
    return pd.read_csv(StringIO(csv.decode()), index_col=0)


def convert_dataframe2l(csv: bytes) -> pd.DataFrame:
    return pd.read_csv(StringIO(csv.decode()), index_col=0, header=[0, 1], parse_dates=True)


def convert_json(val):
    """Convert json string to object."""
    return json.loads(val.decode()) if val is not None else None


sqlite3.register_converter("date", convert_date)
sqlite3.register_converter("datetime", convert_datetime)
sqlite3.register_converter("timestamp", convert_timestamp)
sqlite3.register_converter("json", convert_json)
sqlite3.register_converter("text_series", convert_dataframe)
sqlite3.register_converter("text_quotedata", convert_dataframe2l)


def serialize_to_db(val):
    if val is None:
        return None
    elif isinstance(val, (int, float, bool, str, datetime.datetime)):
        return val
    elif isinstance(val, (pd.DataFrame, pd.Series)):
        return adapt_dataframe(val)
    elif isinstance(val, bytes):
        return adapt_bytes(val)
    else:
        return adapt_json(val)


class MTDB:

    thread_local = threading.local()

    @staticmethod
    def conn():
        ''' Return a thread local db connection. 
            Reference: https://ricardoanderegg.com/posts/python-sqlite-thread-safety/ 
        '''
        try:
            return MTDB.thread_local.db
        except Exception:
            if 'pytest' not in sys.modules:
                MTDB.thread_local.db = sqlite3.connect(
                    os.path.expanduser('~/.minitrade/database/minitrade.db'),
                    detect_types=sqlite3.PARSE_DECLTYPES | sqlite3.PARSE_COLNAMES)
            else:
                MTDB.thread_local.db = sqlite3.connect(
                    os.path.expanduser('~/.minitrade/database/minitrade.pytest.db'),
                    detect_types=sqlite3.PARSE_DECLTYPES | sqlite3.PARSE_COLNAMES)
            MTDB.thread_local.db.row_factory = sqlite3.Row
            return MTDB.thread_local.db

    @staticmethod
    def get_one(tablename: str, where_key: str, where_value: Any, *, cls=None):
        '''Return single row matching the query or None if no rows are available.'''
        table = Table(tablename)
        stmt = Query.from_(table).select('*')
        if where_key is not None:
            if where_value is None:
                stmt = stmt.where(table[where_key].isnull())
            else:
                stmt = stmt.where(table[where_key] == where_value)
        with MTDB.conn() as conn:
            row = conn.execute(str(stmt)).fetchone()
        return row if cls is None or row is None else cls(**row)

    @staticmethod
    def get_all(tablename: str, where_key: str = None, where_value: Any = None, *, where: dict = None, orderby: str |
                tuple[str, bool] = None, limit: int = None, cls=None):
        '''Return all rows matching the query as a list, or empty list if no rows are available.'''
        table = Table(tablename)
        stmt = Query.from_(table).select('*')
        if where_key is not None:
            if where_value is None:
                stmt = stmt.where(table[where_key].isnull())
            else:
                stmt = stmt.where(table[where_key] == where_value)
        if where:
            for k, v in where.items():
                if v is None:
                    stmt = stmt.where(table[k].isnull())
                else:
                    stmt = stmt.where(table[k] == v)
        if orderby is not None:
            try:
                field, ascending = orderby
                ascending = Order.asc if ascending else Order.desc
            except Exception:
                field, ascending = orderby, Order.asc
            stmt = stmt.orderby(field, order=ascending)
        if limit:
            stmt = stmt.limit(limit)
        with MTDB.conn() as conn:
            rows = conn.execute(str(stmt)).fetchall()
        return rows if cls is None else [cls(**r) for r in rows]

    @staticmethod
    def update(tablename: str, where_key: str, where_value: Any, *, values: dict):
        '''Update rows matching the query'''
        table = Table(tablename)
        stmt = Query.update(table)
        if where_key is not None:
            if where_value is None:
                stmt = stmt.where(table[where_key].isnull())
            else:
                stmt = stmt.where(table[where_key] == where_value)
        for k, v in values.items():
            stmt = stmt.set(table[k], serialize_to_db(v))
        with MTDB.conn() as conn:
            conn.execute(str(stmt))

    @staticmethod
    def save(tablename: str, objects, *, on_conflict='error', whitelist=None, blacklist=None):
        '''Save objects to ddb'''
        if not isinstance(objects, list):
            objects = [objects]
        for obj in objects:
            data = obj if isinstance(obj, dict) else asdict(obj)
            if whitelist:
                whitelisted = {k: data.get(k, None) for k in whitelist}
                diff = set(data.keys()) - set(whitelisted.keys())
                if diff:
                    logger.warning(f'Unknonw attributes {diff} detected in {data}, not saved to {tablename}')
                data = whitelisted
            if blacklist:
                data = {k: v for k, v in data.items() if k not in blacklist}
            data = {k: serialize_to_db(v) for k, v in data.items()}
            table = Table(tablename)
            if on_conflict == 'error':
                stmt = Query.into(table).columns(*data.keys()).insert(*data.values())
            elif on_conflict == 'update':
                stmt = Query.into(table).columns(*data.keys()).replace(*data.values())
            elif on_conflict == 'ignore':
                stmt = Query.into(table).columns(*data.keys()).insert(*data.values()).ignore()
                stmt = str(stmt).replace('INSERT IGNORE', 'INSERT OR IGNORE')   # patch for sqlite syntax
            else:
                raise ValueError(f'Unknown conflict handling: {on_conflict}')
            with MTDB.conn() as conn:
                conn.execute(str(stmt))

    @staticmethod
    def delete(tablename: str, where_key: str, where_value: Any):
        '''Delete rows matching the query'''
        table = Table(tablename)
        stmt = Query.from_(table).delete()
        if where_key is not None:
            if where_value is None:
                stmt = stmt.where(table[where_key].isnull())
            else:
                stmt = stmt.where(table[where_key] == where_value)
        with MTDB.conn() as conn:
            conn.execute(str(stmt))

    @staticmethod
    def uniqueid():
        ''' Generate a short unique ID '''
        return generate('1234567890abcdef', 20)





from twelvedata import TDClient

from minitrade.datasource.base import QuoteSource
from minitrade.utils.config import config


class TwelveDataQuoteSource(QuoteSource):
    '''TwelveDataQuoteSource retrieves data from https://twelvedata.com. (Experimental)
    '''

    def __init__(self, api_key: str = None) -> None:
        super().__init__()
        api_key = api_key or config.sources.twelvedata.api_key
        if not api_key:
            raise AttributeError('TwelveData API key is not configured')
        self.client = TDClient(apikey=api_key)

    def _format_ticker(self, ticker):
        return ticker

    def _ticker_timezone(self, ticker):
        # TODO: get timezone from API
        return 'America/New_York'

    def _ticker_calendar(self, ticker):
        # TODO: get market calendar from API
        return 'NYSE'

    def _daily_bar(self, ticker, start, end):
        df = self.client.time_series(
            symbol=self._format_ticker(ticker),
            interval='1day',
            start_date=start,
            end_date=end,
            outputsize=5000,
            timezone='Exchange',
            order='ASC',
        ).as_pandas()
        df.rename(columns={'open': 'Open', 'high': 'High', 'low': 'Low',
                  'close': 'Close', 'volume': 'Volume'}, inplace=True)
        df.index.rename('Date', inplace=True)
        df.index = df.index.tz_localize(None).normalize()
        return df

    def _minute_bar(self, ticker: str, start: str, end: str, interval: int):
        raise NotImplementedError()

    def _spot(self, tickers):
        raise NotImplementedError()




from datetime import datetime, timezone

import akshare as ak
import pandas as pd

from minitrade.datasource.base import QuoteSource

# Need to manually copy libmini_racer.dylib on Mac M1, see https://github.com/sqreen/PyMiniRacer/issues/143


class EastMoneyQuoteSource(QuoteSource):
    '''EastMoneyQuoteSource retrieves data from https://www.eastmoney.com. (Experimental)
    '''

    def _ticker_timezone(self, ticker):
        return 'Asia/Shanghai'

    def _ticker_calendar(self, ticker):
        return 'SSE'

    def _daily_bar(self, ticker, start, end):
        try:
            df: pd.DataFrame = ak.stock_zh_a_hist(
                symbol=ticker, period="daily",
                start_date=start.replace('-', ''),
                end_date=end.replace('-', '') if end else '20500101',   # magic number used in ak lib
                adjust='qfq')
        except KeyError:
            df: pd.DataFrame = ak.fund_etf_hist_em(
                symbol=ticker, period="daily",
                start_date=start.replace('-', ''),
                end_date=end.replace('-', '') if end else '20500101',   # magic number used in ak lib
                adjust='qfq')
        df = df.rename(columns={'日期': 'Date', '开盘': 'Open', '收盘': 'Close', '最高': 'High', '最低': 'Low', '成交量': 'Volume'})
        df['Date'] = pd.to_datetime(df['Date'], format='%Y-%m-%d')
        df = df.set_index('Date')
        df = df[['Open', 'High', 'Low', 'Close', 'Volume']]
        return df

    def _minute_bar(self, ticker: str, start: str, end: str, interval: int):
        raise NotImplementedError()

    def _spot(self, tickers):
        try:
            src = [ak.stock_zh_a_spot_em, ak.fund_etf_spot_em, ak.fund_lof_spot_em]
            df = [f().rename(columns={'代码': 'ticker', '最新价': 'price'}).set_index('ticker') for f in src]
            s = pd.concat(df)['price']
            data = {ticker: s[ticker] if ticker in s.index else None for ticker in tickers}
            return pd.Series(data, name=datetime.now(timezone.utc)).astype(float)
        except Exception as e:
            raise AttributeError(f'Data error') from e





import warnings

from minitrade.broker.base import Broker, BrokerAccount
from minitrade.datasource.base import QuoteSource
from minitrade.utils.config import config


class InteractiveBrokersQuoteSource(QuoteSource):
    '''InteractiveBrokersQuoteSource retrieves data from InteractiveBrokers and requires a broker account to access its service.

Accepted symbol format:
- To precisely identify a ticker, you should use the IBKR contract ID, e.g. "756733" for SPDR S&P 500 ETF.
- As a convenience, you can also use the ticker symbol, e.g. "SPY" for SPDR S&P 500 ETF. However, the resolution from ticker symbol to contract ID can be ambiguous. If multiple contract IDs are found for the same symbol, the first one, preferrable an U.S. instrument, will be used. For example, "SPY" will be resolved to "756733" (SPY ARCA) instead of "237937002" (SPY ASX). If you want to resolve a ticker symbol to contract ID manually, use the `lookup()` method.

Daily bar:
- The daily OHLCV data returns historical data up to T-1.

Minute bar:
- Not available

Spot price:
- Experimental
    '''

    def __init__(self, alias: str | None = None) -> None:
        super().__init__()
        alias = alias or config.sources.ib.account
        if not alias:
            raise ValueError('No IB account specified')
        self.account = BrokerAccount.get_account(alias)
        self.broker = Broker.get_broker(self.account)

    def _format_ticker(self, ticker):
        if ticker.isnumeric():
            return ticker
        else:
            resolved = self.broker.resolve_tickers(ticker)[ticker]
            if len(resolved) == 0:
                raise ValueError(f'Unknown ticker {ticker}')
            if len(resolved) > 1:
                warnings.warn(f'Ambiguous ticker {ticker} resolved to {resolved[0]["label"]}')
            return resolved[0]['id']

    def _ticker_timezone(self, ticker):
        raise NotImplementedError()

    def _ticker_calendar(self, ticker):
        raise NotImplementedError()

    def _daily_bar(self, ticker, start, end):
        self.broker.connect()
        return self.broker.daily_bar(self._format_ticker(ticker), start, end)

    def _minute_bar(self, ticker: str, start: str, end: str, interval: int):
        raise NotImplementedError()

    def _spot(self, tickers):
        self.broker.connect()
        return self.broker.spot([self._format_ticker(ticker) for ticker in tickers])

    def lookup(self, ticker_css):
        self.broker.connect()
        return self.broker.resolve_tickers(ticker_css)



import os
from contextlib import closing
from importlib.metadata import version

import pandas as pd
import requests
import streamlit as st

from minitrade.utils.config import config
from minitrade.utils.mtdb import MTDB

st.set_page_config(page_title='Minitrade Status', layout='wide')


def __call_scheduler(method: str, path: str, params: dict | None = None):
    url = f'http://{config.scheduler.host}:{config.scheduler.port}{path}'
    resp = requests.request(method=method, url=url, params=params)
    if resp.status_code == 200:
        return resp.json()
    elif resp.status_code >= 400:
        raise RuntimeError(f'Scheduler {method} {url} {params} returns {resp.status_code} {resp.text}')


def __call_ibgateway_admin(method: str, path: str, params: dict | None = None):
    '''Call the ibgateway's admin API'''
    url = f'http://{config.brokers.ib.gateway_admin_host}:{config.brokers.ib.gateway_admin_port}{path}'
    resp = requests.request(method=method, url=url, params=params)
    if resp.status_code == 200:
        return resp.json()
    elif resp.status_code >= 400:
        raise RuntimeError(f'Request {path} returned {resp.status_code} {resp.text}')


def display_process_status():
    st.caption('Processes')
    cmd = 'ps -o pid,ppid,command | grep python | grep minitrade | grep -v grep'
    output = os.popen(cmd).read()
    pid_map = {pid: name for pid, ppid, _, _, name, *_ in [line.split() for line in output.splitlines()] if ppid == '1'}
    proc_map = {'scheduler': [], 'ib': [], 'web': []}
    for k, v in pid_map.items():
        proc_map[v] = proc_map.get(v, []) + [k]
    ppid_map = {pid: ppid for pid, ppid, _, _, _, *_ in [line.split() for line in output.splitlines()]}
    for k, v in proc_map.items():
        if len(v) == 0:
            st.warning(f'Process "{k}" is not running.')
        elif len(v) == 1:
            child_pid = [i for i, j in ppid_map.items() if j == v[0]]
            st.text(f'Process "{k}" is running, pid {v}' + (f', child pid {child_pid}.' if len(child_pid) > 0 else '.'))
        else:
            st.error(
                f'Process "{k}" is running, {len(v)} instances found {v}, which is not right. Consider restarting the service.')


def display_trade_status():
    st.caption(f'Scheduled Trade Plans')
    try:
        data = __call_scheduler('GET', '/strategy')
        if len(data) == 0:
            st.text('No trade plan is scheduled.')
        else:
            st.write(pd.DataFrame(data))
    except Exception as e:
        st.error(f'Failed to get trade plan status. {e}')


def display_task_status():
    st.caption(f'Scheduled Tasks')
    try:
        data = __call_scheduler('GET', '/task')
        if len(data) == 0:
            st.text('No task is scheduled.')
        else:
            st.write(pd.DataFrame(data))
    except Exception as e:
        st.error(f'Failed to get task status. Try refresh.')


def display_ib_gateway_status():
    try:
        st.caption('IB Gateways')
        data = __call_ibgateway_admin('GET', '/ibgateway')
        if len(data) == 0:
            st.text('No gateway is running.')
        else:
            st.write(pd.DataFrame(data))
    except Exception as e:
        st.error(f'Failed to get IB gateway status: {e}')


def display_minitrade_version():
    st.caption('Version')
    st.text(version('minitrade'))


def display_database_status():
    tables = [
        'TraderLog',
        'BacktestLog',
        'OrderValidatorLog',
        'ManualTrade',
        'IbOrderLog',
        'IbOrder',
        'IbTrade',
        'RawOrder',
        'TaskLog',
    ]
    st.caption('Database')
    with closing(MTDB.conn()) as conn:
        stats = {
            table:
            (conn.execute(f'select count(*) from {table}').fetchone()[0],
             conn.execute(f'SELECT SUM("pgsize")/1024.0/1024.0 FROM "dbstat" WHERE name="{table}"').fetchone()[0])
            for table in tables}
        st.write(pd.DataFrame(stats, index=['Row', 'Size (MB)']))


st.button('Refresh', type='primary')
st.header('Minitrade Status')

for func in [display_minitrade_version, display_process_status, display_trade_status, display_task_status,
             display_ib_gateway_status, display_database_status]:
    st.text("")
    func()
    st.text("")



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime

from minitrade.datasource.base import QuoteSource
from minitrade.utils.mtdb import MTDB


class UnionQuoteSource(QuoteSource):
    '''UnionQuoteSource allows retrieving data from several underlying data sources. One can configure, for each ticker, the data source and parameters to use for retrieving data.

The data source is configured in the following format:

    [(ticker, source, params), ...]

Where:
- ticker: the ticker or comma separated list of tickers to retrieve data for.
- source: the data source name as accepted by the `QuoteSource.get_source()`.
- params: the additonal parameters to pass to `QuoteSource.get_source()`.

Example:
```
data_source = QuoteSource.get_source('Union', config = [
    ('756733', 'IB', {'alias': 'my_ib_account}),                        # Retrieve SPY from Interactive Brokers ('756733' is the contract id for SPY)
    ('VX', 'CboeFutures', {'cf_method': 'backward', 'roll_day': -7}),   # Retrieve VIX futures from Cboe Futures
    ('AAPL,MSFT', 'Yahoo', None),                                       # Retrieve AAPL and MSFT from Yahoo
    (None, 'Yahoo', None)                                               # Retrieve all other tickers from Yahoo
])
```

Accepted symbol format:
- Same as the underlying data source.

Daily bar:
- Data recency is determined by the most recent data source. For example, if source A provides data up to yesterday (T-1) and source B provides data up to today (T), the combined data from sources A and B will be up to today (T), while A's data at time T is forward filled by data at T-1.

Minute bar:
- Not available

Spot price:
- Not available
    '''

    def __init__(self, config: list = []) -> None:
        super().__init__()
        self.source_map = {}
        for tickers, name, params in config:
            source = QuoteSource.get_source(name, **(params or {}))
            if tickers:
                self.source_map |= {ticker: source for ticker in tickers.split(',')}
            elif None not in self.source_map:
                self.source_map[None] = source
            else:
                raise ValueError('Only one fallback data source is allowed.')

    def _format_ticker(self, ticker):
        return ticker

    def _ticker_timezone(self, ticker):
        if ticker in self.source_map:
            return self.source_map[ticker].ticker_timezone(ticker)
        elif None in self.source_map:
            return self.source_map[None].ticker_timezone(ticker)
        else:
            raise ValueError(f'Ticker {ticker} is not found.')

    def _ticker_calendar(self, ticker):
        if ticker in self.source_map:
            return self.source_map[ticker].ticker_calendar(ticker)
        elif None in self.source_map:
            return self.source_map[None].ticker_calendar(ticker)
        else:
            raise ValueError(f'Ticker {ticker} is not found.')

    def _daily_bar(self, ticker, start, end):
        if ticker in self.source_map:
            return self.source_map[ticker].daily_bar(ticker, start, end)[ticker]
        elif None in self.source_map:
            return self.source_map[None].daily_bar(ticker, start, end)[ticker]
        else:
            raise ValueError(f'Ticker {ticker} is not found.')

    def _minute_bar(self, ticker, start, end, interval):
        raise NotImplementedError()

    def _spot(self, tickers):
        raise NotImplementedError()


@dataclass(kw_only=True)
class UnionQuoteSourceConfig:

    name: str
    '''Data source name'''

    config: list
    '''Configuration in the format of [(ticker, source, params), ...]'''

    update_time: datetime
    '''Time when the union source is updated'''

    @staticmethod
    def list() -> list[UnionQuoteSourceConfig]:
        '''Return the list of available union source configs

        Returns:
            A list of zero or more union source configs
        '''
        configs = MTDB.get_all('UnionQuoteSourceConfig', orderby='name', cls=UnionQuoteSourceConfig)
        return [x.name for x in configs if x.name not in QuoteSource.SYSTEM_SOURCES]

    @staticmethod
    def get_config(name: str) -> UnionQuoteSourceConfig:
        '''Look up a union source config by name

        Args:
            name: Union name

        Returns:
            Union source config if found or None
        '''
        return MTDB.get_one('UnionQuoteSourceConfig', 'name', name, cls=UnionQuoteSourceConfig)

    def save(self) -> None:
        '''Save or update a union source config to database.
        '''
        MTDB.save('UnionQuoteSourceConfig', self, on_conflict='update')

    def delete(self) -> None:
        '''Delete a union source config from database.
        '''
        MTDB.delete('UnionQuoteSourceConfig', 'name', self.name)




import itertools
import json
from datetime import datetime

import pandas as pd
import streamlit as st
from matplotlib import pyplot as plt

from minitrade.broker.base import BrokerAccount
from minitrade.datasource import QuoteSource
from minitrade.datasource.union import UnionQuoteSource, UnionQuoteSourceConfig
from minitrade.utils.config import config

st.set_page_config(page_title='Datasource', layout='wide')


def test_and_save_yahoo_proxy(proxy):
    st.caption('Getting SPY')
    df = QuoteSource.get_source('Yahoo', proxy=proxy).daily_bar('SPY', start='2023-01-01')
    if len(df) > 0:
        st.write(df)
        config.sources.yahoo.proxy = proxy
        config.save()
        st.success('Setting saved')
    else:
        st.error('Getting data from Yahoo not working, please check proxy setting')


def test_and_save_eodhd_api_key(api_key):
    st.caption('Getting SPY')
    df = QuoteSource.get_source('EODHistoricalData', api_key=api_key).daily_bar('SPY', start='2023-01-01')
    if len(df) > 0:
        st.write(df)
        config.sources.eodhd.api_key = api_key
        config.save()
        st.success('Setting saved')
    else:
        st.error('Getting data from EODHistoricalData not working, please check API key')


def test_and_save_twelvedata_api_key(api_key):
    st.caption('Getting SPY')
    df = QuoteSource.get_source('TwelveData', api_key=api_key).daily_bar('SPY', start='2023-01-01')
    if len(df) > 0:
        st.write(df)
        config.sources.twelvedata.api_key = api_key
        config.save()
        st.success('Setting saved')
    else:
        st.error('Getting data from TwelveData not working, please check API key')


def test_and_save_alpaca_api_key(api_key, api_secret):
    st.caption('Getting SPY')
    df = QuoteSource.get_source('Alpaca', api_key=api_key, api_secret=api_secret).daily_bar('SPY', start='2023-01-01')
    if len(df) > 0:
        st.write(df)
        config.sources.alpaca.api_key = api_key
        config.sources.alpaca.api_secret = api_secret
        config.save()
        st.success('Setting saved')
    else:
        st.error('Getting data from Alpaca not working, please check API key')


def test_and_save_tiingo_api_key(api_key):
    st.caption('Getting SPY')
    df = QuoteSource.get_source('Tiingo', api_key=api_key).daily_bar('SPY', start='2023-01-01')
    if len(df) > 0:
        st.write(df)
        config.sources.tiingo.api_key = api_key
        config.save()
        st.success('Setting saved')
    else:
        st.error('Getting data from Tiingo not working, please check API key')


def test_and_save_ib_account(alias):
    st.caption('Getting SPY')
    df = QuoteSource.get_source('InteractiveBrokers', alias=alias).daily_bar('SPY', start='2023-01-01')
    if len(df) > 0:
        st.write(df)
        config.sources.ib.account = alias
        config.save()
        st.success('Setting saved')
    else:
        st.error('Getting data from InteractiveBrokers not working, please check account')


def confirm_delete_union_data_source(source) -> None:
    def confirm_delete():
        if st.session_state.delete_confirm_textinput == source:
            UnionQuoteSourceConfig.get_config(source).delete()
    st.text_input(f'Type "{source}" and press Enter to delete',
                  on_change=confirm_delete, key='delete_confirm_textinput')


def show_source_page():
    source = st.sidebar.radio('Source', QuoteSource.list())

    c1, c2 = st.columns([5, 1], gap='large')
    c1.subheader(source)
    if source in QuoteSource.list():
        if c2.button('Delete', type='primary'):
            confirm_delete_union_data_source(source)
    t1, t2 = st.tabs(['Config', 'About'])
    with t1:
        config_source(source)
    with t2:
        about_source(source)


def about_source(source):
    if source in QuoteSource.list():
        st.markdown(QuoteSource.get_source(source).__doc__ or '')
    else:
        st.write('Unknown data source')


def config_source(source):
    if source == 'Yahoo':
        proxy = st.text_input('HTTP Proxy (Socks proxy not supported)',
                              placeholder='http://host:port', value=config.sources.yahoo.proxy or '') or None
        if st.button('Test and save'):
            test_and_save_yahoo_proxy(proxy)
    elif source == 'EastMoney':
        st.write('Nothing to configure')
    elif source == 'EODHistoricalData':
        api_key = st.text_input('API Key', value=config.sources.eodhd.api_key or '') or None
        if st.button('Test and save'):
            test_and_save_eodhd_api_key(api_key)
    elif source == 'TwelveData':
        api_key = st.text_input('API Key', value=config.sources.twelvedata.api_key or '') or None
        if st.button('Test and save'):
            test_and_save_twelvedata_api_key(api_key)
    elif source == 'Alpaca':
        api_key = st.text_input('API Key', value=config.sources.alpaca.api_key or '') or None
        api_secret = st.text_input('API Secret', value=config.sources.alpaca.api_secret or '') or None
        if st.button('Test and save'):
            test_and_save_alpaca_api_key(api_key, api_secret)
    elif source == 'Tiingo':
        api_key = st.text_input('API Key', value=config.sources.tiingo.api_key or '') or None
        if st.button('Test and save'):
            test_and_save_tiingo_api_key(api_key)
    elif source == 'InteractiveBrokers':
        accounts = [_ for _ in BrokerAccount.list() if _.broker == 'IB']
        if len(accounts) == 0:
            st.warning('Please make sure you have configured your IB account in the broker section.')
            return
        alias = st.selectbox('Use for data access', [_.alias for _ in accounts])
        if st.button('Save'):
            config.sources.ib.account = alias
            config.save()
            st.success('Setting saved')
        if config.sources.ib.account:
            st.success(f'Currently using account "**{config.sources.ib.account}**" for data access.')
        else:
            st.error('No account is currently configured for data access.')
    elif source == 'CboeIndex':
        st.write('Nothing to configure')
    elif source == 'CboeFutures':
        st.write('Nothing to configure')
    elif source in UnionQuoteSource.list():
        st.write('This is an union data source configured as follows:')
        union_config = UnionQuoteSourceConfig.get_config(source).config
        for ticker, name, params in union_config:
            st.caption(ticker or 'Catch-all')
            if params:
                params = ', '.join([f'{k}={v}' for k, v in params.items()])
                st.code(f'QuoteSource.get_source("{name}", {params})')
            else:
                st.code(f'QuoteSource.get_source("{name}")')
    else:
        st.write('Unknown data source')


@st.cache_data(ttl='1h')
def read_daily_bar(ticker, start, end, sources):
    data = {}
    for s in sources:
        try:
            data[s] = QuoteSource.get_source(s).daily_bar(ticker, start=start, end=end)
        except Exception as e:
            raise RuntimeError(f'Getting {s} data error: {e}')
    return data


def compare_data_from_sources(data, s1, s2, tab):
    df1 = data[s1].droplevel(0, axis=1)
    df2 = data[s2].droplevel(0, axis=1)
    df = pd.merge(df1, df2, how='outer', left_index=True, right_index=True, suffixes=('_1', '_2'))
    cols = tab.columns(5)
    for col, c in zip(cols, ['Open', 'High', 'Low', 'Close', 'Volume']):
        col.write(c)
        res = calc_col_diff(s1, s2, df, c)
        col.dataframe(res)
        # plot s1 vs s2
        fig = plot_price(s1, s2, c, res)
        col.pyplot(fig)
        plt.close(fig)
        # print diff stats
        col.write('Diff stats')
        stats = res['diff (%)'].describe()[['count', 'mean', 'std', 'min', 'max']].to_frame().T
        col.write(stats)
        # print top 3 diff
        col.write('Top 3 diff')
        col.write(res.sort_values('diff (%)', ascending=False, key=abs).head(3))
        # plot diff
        fig = plot_diff(c, res)
        col.pyplot(fig)
        plt.close(fig)


def plot_diff(c, res):
    fig, ax = plt.subplots()
    ax.xaxis.set_tick_params(labelbottom=False)
    ax.set_xticks([])
    ax.plot(res.index, res['diff (%)'])
    ax.set_title(f'{c} diff (%)')
    ax.set_xlabel('Date')
    ax.set_ylabel('Diff (%)')
    fig.set_size_inches(6, 3)
    return fig


def plot_price(s1, s2, c, res):
    fig, ax = plt.subplots()
    ax.xaxis.set_tick_params(labelbottom=False)
    ax.set_xticks([])
    ax.plot(res.index, res[s1], label=s1)
    ax.plot(res.index, res[s2], label=s2)
    ax.legend()
    ax.set_title(f'{c} {s1} vs {s2}')
    ax.set_xlabel('Date')
    ax.set_ylabel(c)
    fig.set_size_inches(6, 3)
    return fig


@st.cache_data(ttl='1h')
def calc_col_diff(s1, s2, df, c):
    df[f'{c} diff'] = df[f'{c}_2'] - df[f'{c}_1']
    df[f'{c} diff (%)'] = df[f'{c} diff'] / df[f'{c}_1'] * 100
    res = df[[f'{c}_1', f'{c}_2', f'{c} diff', f'{c} diff (%)']]
    res.columns = [s1, s2, 'diff', 'diff (%)']
    res.index = res.index.strftime('%Y-%m-%d')
    return res


def plot_ohlcv(data):
    for ticker in data.columns.get_level_values(0).unique():
        c1, c2, c3 = st.columns(3)
        df = data[ticker]
        # print ohlcv
        c1.write(f'### {ticker}')
        c1.dataframe(df)
        c1.download_button(f'Download {ticker}', df.to_csv().encode('utf-8'), f'{ticker}.csv')
        # plot stats
        c2.write('### Stats')
        stats = df.describe()
        c2.dataframe(stats)
        # plot ohlcv
        fig, ax = plt.subplots()
        ax.xaxis.set_tick_params(labelbottom=False)
        ax.set_xticks([])
        ax.plot(df.index, df['Open'], label='Open')
        ax.plot(df.index, df['High'], label='High')
        ax.plot(df.index, df['Low'], label='Low')
        ax.plot(df.index, df['Close'], label='Close')
        ax.legend()
        ax.set_title(f'{ticker}')
        ax.set_xlabel('Date')
        ax.set_ylabel('Price')
        # plot volume
        ax2 = ax.twinx()
        ax2.bar(df.index, df['Volume'], label='Volume', alpha=0.3)
        ax2.set_ylabel('Volume')
        ax2.legend(loc='upper right')
        ax2.grid(False)
        fig.set_size_inches(8, 4)
        c3.write('### Chart')
        c3.pyplot(fig)
        plt.close(fig)
        st.write('---')


def inspect_data():
    tickers = st.sidebar.text_input('Ticker', value='SPY')
    start = st.sidebar.text_input('Start date', value='2000-01-01') or None
    end = st.sidebar.text_input('End date') or None
    sources = st.sidebar.multiselect('Sources', QuoteSource.list(), default=['Yahoo'])

    if len(sources) == 1 and st.sidebar.button('Inspect'):
        if len(sources) == 1:
            st.write(f'## {sources[0]}')
            try:
                df = read_daily_bar(tickers, start, end, sources)[sources[0]]
                plot_ohlcv(df)
            except Exception as e:
                st.error(e)
    if len(sources) > 1 and st.sidebar.button('Compare'):
        if len(tickers.split(',')) > 1:
            st.sidebar.warning('Only the first ticker will be compared.')
        tickers = tickers.split(',')[0]
        st.write(f'## {tickers}')
        data = read_daily_bar(tickers, start, end, sources)
        pairs = list(itertools.combinations(sources, 2))
        tabs = st.tabs([(f'{s1} vs {s2}') for s1, s2 in pairs])
        for i, (s1, s2) in enumerate(pairs):
            with tabs[i]:
                compare_data_from_sources(data, s1, s2, tabs[i])


def lookup_ticker():
    if not config.sources.ib.account:
        st.sidebar.warning('Please make sure you have configured the IB account for data access.')
        return
    source = QuoteSource.get_source('InteractiveBrokers')
    tickers = st.sidebar.text_input('Lookup IB contract ID', placeholder='SPY,QQQ')
    if st.sidebar.button('Lookup'):
        idmap = source.lookup(tickers)
        st.header('Result')
        for k, idlst in idmap.items():
            with st.expander(k, expanded=True):
                if len(idlst) == 0:
                    st.write('No found')
                else:
                    for item in idlst:
                        st.write(f'{item["id"]} - {item["label"]}')


def create_union_source():
    st.subheader('Create union data source')
    st.info('A union data source let you retrieve data for different tickers using different data sources. '
            'You can explicitly map tickers to data sources. '
            'You can also define a catch-all data source for tickers not found in the mapping.')

    source_name = st.text_input('Data source name')
    with st.expander('Sources', expanded=True):
        tabs = st.tabs('12345678')
        ticker = [None] * 8
        name = [None] * 8
        params = [None] * 8
        for i, tab in enumerate(tabs):
            with tab:
                ticker[i] = st.text_input('Tickers (comma separated list without space)', key=f'ticker_{i}')
                name[i] = st.selectbox('Source', QuoteSource.SYSTEM_SOURCES, key=f'name_{i}')
                params[i] = st.text_input(
                    'Params (parameters to be passed to `QuoteSource.get_source()` in JSON format)', key=f'params_{i}',
                    placeholder='{}') or None
                try:
                    if params[i]:
                        params[i] = json.loads(params[i])
                except Exception as e:
                    st.error(f'Invalid JSON: {e}')
    with st.expander('Catch-all source', expanded=True):
        catchall = st.selectbox('Source', [None] + QuoteSource.SYSTEM_SOURCES)
        catchall_params = st.text_input('Params') or None
        try:
            if catchall_params:
                json.loads(catchall_params)
        except Exception as e:
            st.error(f'Invalid JSON: {e}')
    save = st.button('Save')
    if save:
        if not source_name:
            st.error('Data source name is required')
            return
        elif source_name in QuoteSource.SYSTEM_SOURCES:
            st.error(f'"{source_name}" is a built-in data source')
            return
        union_config = [(t, n, p) for t, n, p in zip(ticker, name, params) if t and n]
        if catchall:
            union_config.append((None, catchall, catchall_params))
        UnionQuoteSource(union_config)    # test if the config is valid
        union = UnionQuoteSourceConfig(
            name=source_name,
            config=union_config,
            update_time=datetime.utcnow()
        )
        union.save()
        st.success(f'Union data source "{source_name}" saved')


action = st.sidebar.radio('Action', ['Config', 'Create', 'Inspect', 'Lookup (IB only)'])

if action == 'Config':
    show_source_page()


if action == 'Create':
    create_union_source()


if action == 'Inspect':
    inspect_data()


if action == 'Lookup (IB only)':
    lookup_ticker()



from __future__ import annotations

from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor, as_completed
from datetime import datetime, timedelta
from typing import Any
from zoneinfo import ZoneInfo

import pandas as pd
import pandas_market_calendars as mcal

__all__ = [
    'QuoteSource',
]


class QuoteSource(ABC):
    '''
    QuoteSource is a base class that returns quote data for instruments. Extend this class
    to add a concrete implementation to get data from particular data source.

    At a minimum, the following methods should be implemented:
    ```python
    class MyQuoteSource(QuoteSource):
        def _ticker_timezone(self, ticker: str) -> str:
            pass
            
        def _ticker_calendar(self, ticker: str) -> str:
            pass
            
        def _daily_bar(self, ticker: str, start: str, end: str) -> pd.DataFrame:
            pass
    ```
    '''

    SYSTEM_SOURCES = sorted(['Yahoo', 'EODHistoricalData', 'TwelveData', 'Alpaca',
                             'EastMoney', 'Tiingo', 'InteractiveBrokers', 'CboeIndex', 'CboeFutures'])
    '''A list of names for supported quote sources as input to `QuoteSource.get_source()`.'''

    @staticmethod
    def list() -> list[str]:
        '''List all available quote sources.'''
        from minitrade.datasource.union import UnionQuoteSourceConfig
        return QuoteSource.SYSTEM_SOURCES + UnionQuoteSourceConfig.list()

    @staticmethod
    def get_source(name: str, **kwargs: dict[str, Any]) -> QuoteSource:
        '''Get quote source by name

        Args:
            name: Quote source name
            kwargs: Keyword arguments to be passed to the quote source constructor

        Returns:
            A QuoteSource instance

        Raises:
            ValueError: If the asked data source is not supported
        '''
        if name == 'Yahoo':
            from .yahoo import YahooQuoteSource
            return YahooQuoteSource(**kwargs)
        elif name == 'Tiingo':
            from .tiingo import TiingoQuoteSource
            return TiingoQuoteSource(**kwargs)
        elif name == 'EODHistoricalData':
            from .eodhd import EODHistoricalDataQuoteSource
            return EODHistoricalDataQuoteSource(**kwargs)
        elif name == 'TwelveData':
            from .twelve import TwelveDataQuoteSource
            return TwelveDataQuoteSource(**kwargs)
        elif name == 'Alpaca':
            from .alpaca import AlpacaQuoteSource
            return AlpacaQuoteSource(**kwargs)
        elif name == 'EastMoney':
            from .eastmoney import EastMoneyQuoteSource
            return EastMoneyQuoteSource()
        elif name == 'InteractiveBrokers':
            from .ib import InteractiveBrokersQuoteSource
            return InteractiveBrokersQuoteSource(**kwargs)
        elif name == 'CboeIndex':
            from .cboe_index import CboeIndexQuoteSource
            return CboeIndexQuoteSource(**kwargs)
        elif name == 'CboeFutures':
            from .cboe_futures import CboeFuturesQuoteSource
            return CboeFuturesQuoteSource(**kwargs)
        else:
            from .union import UnionQuoteSource, UnionQuoteSourceConfig
            try:
                return UnionQuoteSource(UnionQuoteSourceConfig.get_config(name).config)
            except ValueError:
                raise ValueError(f'Quote source {name} is not supported')

    @abstractmethod
    def _ticker_timezone(self, ticker: str) -> str:
        '''Get the timezone of a ticker.'''
        raise NotImplementedError()

    def ticker_timezone(self, ticker: str) -> str:
        '''Get the timezone of a ticker.'''
        try:
            return self._ticker_timezone(ticker)
        except Exception as e:
            raise RuntimeError(f'Cannot get market timezone for {ticker}') from e

    @abstractmethod
    def _ticker_calendar(self, ticker: str) -> str:
        '''Get the calendar name of a ticker as recognized by `pandas_market_calendars`.'''
        raise NotImplementedError()

    def ticker_calendar(self, ticker: str) -> str:
        '''Get the calendar name of a ticker.'''
        try:
            return self._ticker_calendar(ticker)
        except Exception as e:
            raise RuntimeError(f'Cannot get market calendar for {ticker}') from e

    def today(self, ticker: str) -> datetime:
        '''Get today's date in a ticker's local timezone.'''
        return datetime.now(ZoneInfo(self.ticker_timezone(ticker))).replace(hour=0, minute=0, second=0, microsecond=0)

    def is_trading_now(self, ticker: str) -> bool:
        '''Check if a ticker is trading now.'''
        calendar = mcal.get_calendar(self.ticker_calendar(ticker))
        today = self.today(ticker)
        schedule = calendar.schedule(
            start_date=(today - timedelta(days=30)).strftime('%Y-%m-%d'),
            end_date=(today + timedelta(days=30)).strftime('%Y-%m-%d')
        )
        return calendar.is_open_now(schedule, only_rth=True)

    @abstractmethod
    def _spot(self, tickers: list[str]) -> pd.Series:
        '''Return spot prices for a list of tickers.'''
        raise NotImplementedError()

    def spot(self, tickers: list[str] | str) -> pd.Series:
        '''Read current quote for a list of `tickers`.

        Args:
            tickers: Tickers as a list of string or a comma separated string without space

        Returns:
            Current quotes indexed by ticker as a pandas Series
        '''
        try:
            if isinstance(tickers, str):
                tickers = tickers.split(',')
            return self._spot(tickers)
        except NotImplementedError as e:
            raise e
        except Exception as e:
            raise RuntimeError(e) from e

    @abstractmethod
    def _daily_bar(self, ticker: str, start: str, end: str) -> pd.DataFrame:
        '''Same as `daily_bar()` for only one ticker.

        This should be overridden in subclass to provide an implemention.

        Returns:
            A dataframe with columns 'Open', 'High', 'Low', 'Close', 'Volume' indexed by datetime
        '''
        raise NotImplementedError()

    def daily_bar(
            self, tickers: list[str] | str,
            start: str = '2000-01-01', end: str = None, align: bool = True, normalize: bool = False,
            num_workers: int = 1) -> pd.DataFrame:
        '''Read end-of-day OHLCV data for a list of `tickers` starting from `start` date and ending on `end` date (both inclusive).

        Args:
            tickers: Tickers as a list of string or a comma separated string without space
            start: Start date in string format 'YYYY-MM-DD'
            end: End date in string format 'YYYY-MM-DD'
            align: True to align data to start on the same date, i.e. drop leading days when not all tickers have data available.
            normalize: True to normalize the close price on the start date to 1 for all tickers and scale all price data accordingly.
            num_workers: Number of parallel workers to use for fetching data.

        Returns:
            A dataframe with 2-level columns, first level being the tickers, and the second level being columns 'Open', 'High', 'Low', 'Close', 'Volume'. The dataframe is indexed by datetime.
        '''
        def fetch_data(ticker):
            return ticker, self._daily_bar(ticker, start, end)

        try:
            if isinstance(tickers, str):
                tickers = tickers.split(',')

            with ThreadPoolExecutor(max_workers=num_workers) as executor:
                data = {ticker: None for ticker in tickers}
                futures = [executor.submit(fetch_data, ticker) for ticker in tickers]
                for future in as_completed(futures):
                    ticker, result = future.result()
                    data[ticker] = result

            df = pd.concat(data, axis=1).sort_index().ffill()
            if align:
                start_index = df[df.notna().all(axis=1)].index[0]
                df = df.loc[start_index:, :]
                if normalize:
                    ohlc = ['Open', 'High', 'Low', 'Close']
                    for s in tickers:
                        df.loc[:, (s, ohlc)] = df.loc[:, (s, ohlc)] / df[s].loc[start_index, 'Close']
            elif normalize:
                raise ValueError('normalize=True requires align=True')
            assert isinstance(df.index, pd.DatetimeIndex)
            assert df.index.is_monotonic_increasing
            return df
        except NotImplementedError as e:
            raise e
        except Exception as e:
            raise RuntimeError(e) from e

    def monthly_bar(
            self, tickers: list[str] | str, start: str = '2020-01-01', end: str = None, align: bool = True,
            normalize: bool = False, num_workers: int = 1) -> pd.DataFrame:
        ''' Read monthly OHLCV data for a list of `tickers` starting from `start` date and ending on `end` date (both inclusive).

        Args:
            tickers: Tickers as a list of string or a comma separated string without space
            start: Start date in string format 'YYYY-MM'
            end: End date in string format 'YYYY-MM'
            align: True to align data to start on the same date, i.e. drop leading days when not all tickers have data available.
            normalize: True to normalize the close price on the start date to 1 for all tickers and scale all price data accordingly.
            num_workers: Number of parallel workers to use for fetching data.

        Returns:
            A dataframe with 2-level columns, first level being the tickers, and the second level being columns 'Open', 'High', 'Low', 'Close', 'Volume'. The dataframe is indexed by last day of month.
        '''
        try:
            start = pd.offsets.MonthBegin().rollback(pd.to_datetime(start)).strftime('%Y-%m-%d')
            end = (pd.to_datetime(end) + pd.offsets.MonthEnd(1)).strftime('%Y-%m-%d') if end else None
            daily = self.daily_bar(tickers, start, end, align, normalize, num_workers=num_workers)
            monthly = daily.ta.apply(lambda x: x.resample('M').agg({
                'Open': 'first',
                'High': 'max',
                'Low': 'min',
                'Close': 'last',
                'Volume': 'sum',
            }))
            assert isinstance(monthly.index, pd.DatetimeIndex)
            assert monthly.index.is_monotonic_increasing
            return monthly
        except NotImplementedError as e:
            raise e
        except Exception as e:
            raise RuntimeError(e) from e

    @abstractmethod
    def _minute_bar(self, ticker: str, start: str, end: str, interval: int) -> pd.DataFrame:
        '''Same as `minute_bar()` for only one ticker.

        This should be overridden in subclass to provide an implemention.

        Returns:
            A dataframe with columns 'Open', 'High', 'Low', 'Close', 'Volume' indexed by datetime
        '''
        raise NotImplementedError()

    def minute_bar(
            self, tickers: list[str] | str, start: str = None, end: str = None, interval: int = 1,
            num_workers: int = 1) -> pd.DataFrame:
        '''Read minute OHLCV data for a `ticker` starting from `start` date and ending on `end` date (both inclusive).

        Args:
            tickers: Tickers as a list or comma-separated string
            start: Start date in string format 'YYYY-MM-DD'
            end: End date in string format 'YYYY-MM-DD'
            interval: Interval in minutes
            num_workers: Number of parallel workers to use for fetching data.

        Returns:
            A dataframe with columns 'Open', 'High', 'Low', 'Close', 'Volume' indexed by datetime
        '''
        def fetch_data(ticker):
            return ticker, self._minute_bar(ticker, start, end, interval)

        try:
            if isinstance(tickers, str):
                tickers = tickers.split(',')

            with ThreadPoolExecutor(max_workers=num_workers) as executor:
                data = {ticker: None for ticker in tickers}
                futures = [executor.submit(fetch_data, ticker) for ticker in tickers]
                for future in as_completed(futures):
                    ticker, result = future.result()
                    data[ticker] = result

            df = pd.concat(data, axis=1).sort_index().ffill()
            assert isinstance(df.index, pd.DatetimeIndex)
            assert df.index.is_monotonic_increasing
            return df
        except NotImplementedError as e:
            raise e
        except Exception as e:
            raise RuntimeError(e) from e
